# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np

from ...utils import generate_test_photons
from .. import Parallel
from ...optics import FlatDetector as CCD
from ...design import RowlandTorus, RowlandCircleArray

def test_Parallel_numbering():
    '''Test automatic numbering and numbers assigned to photons.
    '''
    pos = [[0, -10.1, -10.1],[0, .1, -10.1],[0, -10.1, .1],[0, .1, .1]]
    detect = Parallel(elem_class=CCD, elem_args={'pixsize': 0.01, 'zoom': 5},
                      elem_pos={'position': pos}, id_col='CCD_ID')

    for i in range(len(pos)):
        assert detect.elements[i].id_num == i

    photons = generate_test_photons(5)
    photons = detect(photons)
    assert np.all(photons['CCD_ID'] == 3.)

def test_Parallel_offset_numbering():
    '''The randomly generated positions might generate overlapping elements,
    but that is OK to test the numbering scheme.'''
    pos = np.random.rand(5, 3) * 50
    detect = Parallel(elem_class=CCD, elem_args={'pixsize': 0.01, 'zoom': 0.1},
                      elem_pos={'position': pos}, id_num_offset=100)

    for i in range(pos.shape[0]):
        assert detect.elements[i].id_num == i + 100

def test_parallel_calculated_normals():
    '''Regression test: At one point the rotation matrix was reversed,
    mixing active and passive rotations. As a result, the detector elements
    did not match up on the Rowland circle.'''
    rowland = RowlandTorus(5900., 6020.)
    detccdargs = {'pixsize': 0.024,'zoom': [1, 24.576, 12.288]}
    det = RowlandCircleArray(rowland=rowland, elem_class=CCD,
                             elem_args=detccdargs,
                             d_element=49.652, theta=[np.pi - 0.2, np.pi -0.19])
    e1 = det.elements[0]
    e2 = det.elements[1]
    # These two edges should be close together:
    edge1 = e1.geometry('center') + e1.geometry('v_y')
    edge2 = e2.geometry('center') - e2.geometry('v_y')
    assert np.all(np.abs(edge1 - edge2) < 3.)



import numpy as np
import matplotlib.pyplot as plt
from astropy.table import Table
from marxs.source import Source
en = np.arange(0.5, 7., .5)
flux = en**(-2)

# Input as dictionary
dictspectrum = {'energy': en, 'flux': flux}
s1 = Source(energy=dictspectrum, name='dict')

# Input as astropy Table
tablespectrum = Table([en, flux], names=['energy', 'flux'])
s2 = Source(energy=tablespectrum, name='table')

# Input as numpy arrays
numpyspectrum = np.vstack([en, flux])
s3 = Source(energy=numpyspectrum, name='2d array')

fig = plt.figure()
for s in [s1, s2, s3]:
    photons = s.generate_photons(1e3)
    plt.hist(photons['energy'], histtype='step', label=s.name, bins=20)

leg = plt.legend()
lab = plt.xlabel('Energy [keV]')
lab = plt.ylabel('Counts / bin')



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from scipy.stats import normaltest

from ..scatter import RadialMirrorScatter
from ..detector import FlatDetector
from ...utils import generate_test_photons

def test_distribution_of_scattered_rays():
    '''Check that scattered rays have a normal distribution.'''
    photons = generate_test_photons(500)
    photons['pos'] = np.tile(np.array([0., 1., 0., 1.]), (500, 1))
    rms = RadialMirrorScatter(inplanescatter=0.1)
    det = FlatDetector(position=[-1, 0, 0], zoom=1000)

    p = rms(photons)
    p = det(p)

    assert np.all(p['det_y'] == 0)
    stat, pval = normaltest(p['det_x'])
    assert pval > 0.5
    assert np.allclose(np.std(p['det_x']), np.arctan(0.1), rtol=0.1)

def test_distribution_scatered_rays_turned():
    '''Test that the in-plane scattering plane is correctly calculated.

    The photons are positioned 90 deg from the example above.
    Also, do some perpendicular to the plan scattering, that's missing
    from the test above.'''
    photons = generate_test_photons(500)
    photons['pos'] = np.tile(np.array([0., 0., 1., 1.]), (500, 1))
    rms = RadialMirrorScatter(inplanescatter=0.1, perpplanescatter=0.01)
    det = FlatDetector(position=[-1, 0, 0], zoom=1000)

    p = rms(photons)
    p = det(p)

    # This is in-plane-scatter
    assert np.allclose(np.std(p['det_y']), np.arctan(0.1), rtol=0.1)
    # This is scatter perpendicular to the plane.
    assert np.allclose(np.std(p['det_x']), np.arctan(0.01), rtol=0.1)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from scipy.stats import kstest
from astropy.table import Table, Column
from astropy.coordinates import SkyCoord

from transforms3d.axangles import axangle2aff

import pytest

import marxs
import marxs.optics
import marxs.optics.base
import marxs.optics.aperture
import marxs.source
from marxs.utils import generate_test_photons

@pytest.fixture(autouse=True)
def photons1000():
    '''Make a list of photons parallel to optical axis'''
    mysource = marxs.source.PointSource(coords=SkyCoord(30., 30., unit="deg"), energy=1., flux=300.)
    np.random.seed(0)
    p = mysource.generate_photons(1000)
    mypointing = marxs.source.FixedPointing(coords=SkyCoord(30., 30., unit='deg'))
    return mypointing(p)


def test_pos4d_no_transformation():
    '''Test coordinate transforms on initialization of optical elements.

    This test checks that the appropriate transformations are done.
    It should be extended, once derived properties (e.g. a plane constructed
    from transformed points are implemented.
    '''
    oe = marxs.optics.aperture.RectangleAperture()
    assert np.all(oe.pos4d == np.eye(4))
    assert np.all(oe.geometry('center') == np.array([0, 0, 0, 1]))
    assert np.all(oe.geometry('e_y') == np.array([0, 1, 0, 0]))
    assert np.all(oe.geometry('e_z') == np.array([0, 0, 1, 0]))


def test_pos4d_rotation():
    rotation = axangle2aff(np.array([1, 1, 0]), np.deg2rad(90))
    oe = marxs.optics.aperture.RectangleAperture(orientation=rotation[:3,:3])
    assert np.all(oe.pos4d == rotation)
    assert np.all(oe.geometry('center') == np.array([0, 0, 0, 1]))
    assert np.allclose(oe.geometry('e_y'), np.array([.5, .5, 1. / np.sqrt(2), 0]))
    assert np.allclose(oe.geometry('e_z'), np.array([1., -1, 0, 0]) / np.sqrt(2))


def test_pos4d_translation():
    oe = marxs.optics.aperture.RectangleAperture(position = np.array([-2, 3, 4.3]))
    expectedpos4d = np.eye(4)
    expectedpos4d[:3, 3] = np.array([-2, 3, 4.3])
    assert np.all(oe.pos4d == expectedpos4d)
    assert np.all(oe.geometry('center') == np.array([-2, 3, 4.3, 1]))
    assert np.all(oe.geometry('e_y') == np.array([0, 1, 0, 0]))
    assert np.all(oe.geometry('e_z') == np.array([0, 0, 1, 0]))



mark = pytest.mark.parametrize

all_slits = [marxs.optics.aperture.RectangleAperture(zoom=2),
             marxs.optics.aperture.RectangleAperture(zoom=[2,2,2]),
            ]

@mark('myslit', all_slits)
def test_pos4d_transforms_slit(photons1000, myslit):
    '''Test coordinate transforms on initialization of optical elements.

    The initial 4D transforms should be done to any optical element.
    Here, I pick the entrance aperture for testing, because it places the
    positional vector of the plucker coordinates in a plane, independent
    of the initial values.
    '''

    p = myslit(photons1000)
    assert np.allclose(p['pos'][:, 0], 0)
    assert kstest((p['pos'][:, 1] + 2) / 4, "uniform")[1] > 0.01
    assert kstest((p['pos'][:, 2] + 2) / 4, "uniform")[1] > 0.01


def test_pos4d_transforms_slit_rotated(photons1000):
    '''Test coordinate transforms on rotated entrance aperture.'''
    p = photons1000

    rotation = axangle2aff(np.array([0, 1, 0]), np.deg2rad(90))
    myslit = marxs.optics.aperture.RectangleAperture(orientation=rotation[:3, :3], zoom=0.5)
    p = myslit(p)
    assert np.allclose(p['pos'][:, 2], 0)
    assert kstest(p['pos'][:, 0] + 0.5, "uniform")[1] > 0.01
    assert kstest(p['pos'][:, 1] + 0.5, "uniform")[1] > 0.01
    # Sometimes fails, because rays exactly parallel to aperture.
    # Find other example, e.g. 45 deg.
    # Check in aperture if that's the case and raise warning.


def test_photonlocalcoords_decorator():

    pos = np.array([[1, 0, 0, 1], [0, 1, 0, 1]])
    dir = np.array([[1, 0, 0, 0], [0, 1, 0, 0]])
    photontab = Table([Column(data=pos, name='pos'),
                       Column(data=dir, name='dir')
                       ])

    class OEforTest(marxs.optics.base.OpticalElement):
        @marxs.optics.base.photonlocalcoords
        def functiontodecorate(self, photons):
            assert np.allclose(photons['dir'], dir)
            assert np.allclose(photons['pos'], np.array([[2., 2, 3, 1], [1, 3, 3, 1]]))
            return photons

    oetest = OEforTest(position=np.array([-1., -2, -3]))
    photontab = oetest.functiontodecorate(photontab)
    assert np.allclose(photontab['dir'], dir)
    assert np.allclose(photontab['pos'], pos)


def test_FlatStack():
    '''Run a stack of two elements and check that both are applied to the photons.'''
    fs = marxs.optics.FlatStack(position=[0, 5, 0], zoom=2,
                                elements=[marxs.optics.EnergyFilter, marxs.optics.FlatDetector],
                                keywords=[{'filterfunc': lambda x: 0.5},{}])
    # Check all layers are at the same position
    assert np.allclose(fs.geometry('center'), np.array([0, 5, 0, 1]))
    assert np.allclose(fs.elements[0].geometry('center'), np.array([0, 5, 0, 1]))
    assert np.allclose(fs.elements[1].geometry('center'), np.array([0, 5, 0, 1]))
    assert np.allclose(fs.pos4d, fs.elements[1].pos4d)

    p = generate_test_photons(5)
    # Photons 0, 1 miss the stack
    p['pos'][:, 1] = [0, 1, 3.1, 4, 5]
    fs.loc_coos_name = ['a', 'b']
    p = fs(p)
    assert np.allclose(p['probability'], [1, 1, .5, .5, .5])
    assert np.all(np.isnan(p['a'][:2]))
    assert np.allclose(p['a'][2:], [-1.9, -1., 0])



# Licensed under GPL version 3 - see LICENSE.rst
from functools import wraps
from copy import copy

import numpy as np
from astropy.table import Table, Row

from ..math.pluecker import point_dir2plane, dir_point2line, intersect_line_plane
from ..math.utils import e2h, h2e
from ..base import SimulationSequenceElement, _parse_position_keywords

class OpticalElement(SimulationSequenceElement):
    '''Base class for all optical elements in marxs.

    This class cannot be used to instanciate an optical element directly, rather it serves as a
    base class from with other optical elements will be derived.

    At the very minumum, any derived class needs to implement `__call__` which
    typically calls `intersect` and either `process_photon` or
    `process_photons`. If the interaction with the photons (e.g. scattering of
    a mirror surface) can be implemented in a vectorized way using numpy array
    operations, the derived class should overwrite `process_photons`
    (`process_photon` is not used in this case).  If no vectorized
    implementation is available, it is sufficient to overwrite
    `process_photon`.  Marxs will call `process_photons`, which (if not
    overwritten) contains a simple for-loop to loop over all photons in the
    array and call `process_photon` on each of them.

    '''

    def geometry(self, key):
        '''This function wraps access to the pos4d matrix.

        This is mostly a convenience method that gives access to vectors from the
        ``pos4d`` matrix in familiar terms with string labels:

        - ``center``: The ``center`` is the origin of the local coordiante system
          of the optical elemement. Typically, if will be the center of the
          active plane, e.g. the center of the mirror surface.
        - :math:`\hat v_{y,z}`: The box stretches in the y and z direction for
          :math:`\pm \hat v_y` and :math:`\pm \hat v_z`. In optics, the relevant
          interaction often happens on a surface, e.g. the surface of a mirror or
          of a reflection grating. In the defaul configuration, this surface is in
          the yz-plane.
        - :math:`\hat v_x`: The thickness of the box is not as important in many
          cases,
          but is useful to e.g. render the geometry or to test if elements collide.
          The box reaches from the yz-plane down to :math:`- \hat v_x`. Note, that
          this definition means the size parallel to the y and z axis is
          :math:`2 |\hat v_{y,z}|`, but only :math:`1 |\hat v_{x}|`.
        - :math:`\hat e_{x,y,z}`: When an object is initialized, it automatically
          adds unit vectors in the
          direction of :math:`\hat v_{x,y,z}` called :math:`\hat e_{x,y,z}`.
        - ``plane``: It also adds the homogeneous coordinates of the active plane
          as ``plane``.
        - Other labels get looked up in ``self._geometry`` and if the resulting
          value is a 4-d vector, it gets transformed with ``pos4d``.

        Not all these labels make sense for every optical element (e.g. a curved
        mirror does not really have a "plane").
        Access through this method is slower than direct indexing of ``self.pos4d``.
        '''

        if key == 'center':
            return self.pos4d[:, 3]
        elif key in ['v_x', 'e_x']:
            val = self.pos4d[:, 0]
        elif key in ['v_y', 'e_y']:
            val = self.pos4d[:, 1]
        elif key in ['v_z', 'e_z']:
            val = self.pos4d[:, 2]
        elif key == 'plane':
            return point_dir2plane(self.geometry('center'), self.geometry('e_x'))
        else:
            val = self._geometry[key]
            if isinstance(val, np.ndarray) and (val.shape[-1] == 4):
                val = np.dot(self.pos4d, val)
        if key[:2] == 'e_':
            return val / np.linalg.norm(val)
        else:
            return val

    def __init__(self, **kwargs):
        self.pos4d = _parse_position_keywords(kwargs)
        super(OpticalElement, self).__init__(**kwargs)

    def intersect(self, dir, pos):
        '''Calculate the intersection point between a ray and the element

        Parameters
        ----------
        dir : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of the direction of the ray
        pos : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of a point on the ray

        Returns
        -------
        intersect :  boolean array of length N
            ``True`` if an intersection point is found.
        interpos : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of the intersection point. Values are set
            to ``np.nan`` if no intersection point is found.
        interpos_local : `numpy.ndarray` of shape (N, 2)
            y and z coordinates in the coordiante system of the active plane.
        '''
        raise NotImplementedError

    def process_photon(self, dir, pos, energy, polarization):
        '''Simulate interaction of optical element with a single photon.

        This is called from the `process_photons` method in a loop over all
        photons. That method also collects the output values and inserts them
        into the photon list. ``process_photon`` can return any number of
        values in additon to the required dir, pos, etc.. Define a class
        attribute ``output_columns`` as a list of strings to determine how into
        which column these numbers should be inserted.

        Parameters
        ----------
        dir : `numpy.ndarray`
            4-d direction vector of ray in homogeneous coordinates
        pos : `numpy.ndarray`
            4-d position of last interaction pf the photons with any optical element in
            homogeneous coordinates. Together with ``dir`` this determines the equation
            of the ray.
        energy : float
            Photon energy in keV.
        polarization : float
            Polarization angle of the photons.

        Returns
        -------
        dir : `numpy.ndarray`
            4-d direction vector of ray in homogeneous coordinates
        pos : `numpy.ndarray`
            4-d position of last interaction pf the photons with any optical element in
            homogeneous coordinates. Together with ``dir`` this determines the equation
            of the ray.
        energy : float
            Photon energy in keV.
        polarization : float
            Polarization angle of the photons.
        probability : float
            Probability that the photon passes this optical element. Set to 0 if the
            photon is absorbed, to 1 if it passes and to number between 0 and 1 to
            express a probability that the photons passes.
        other : floats
            One number per entry in `output_columns`.
        '''
        raise NotImplementedError
        return dir, pos, energy, polarization, probability, any, other, output, columns

    def __call__(self, photons):
        intersect_out = self.intersect(photons['dir'].data, photons['pos'].data)
        return self.process_photons(photons, *intersect_out)

    def process_photons(self, photons, intersect, interpos, intercoos):
        '''Simulate interaction of optical element with photons - vectorized.

        Derived classes should overwrite this function or `process_photon`.

        Parameters
        ----------
        photons: `astropy.table.Table` or `astropy.table.Row`
            Table with photon properties
        intersect, interpos, intercoos : array (N, 4)
            The array ``interpos`` contains the intersection points in the global
            coordinate system, ``intercoos`` in a local coordiante system (2d in most cases).

        Returns
        -------
        photons: `astropy.table.Table` or `astropy.table.Row`
            Table with photon properties.
            If possible, the input table is modified in place, but in some
            cases this might not be possible and the returned Table may be
            a copy. Do not rely on either - use ``photons.copy()`` if you want
            to ensure you are working with an independent copy.
        '''
        if intersect.sum() > 0:
            self.add_output_cols(photons, self.loc_coos_name)
            # Add ID number to ID col, if requested
            if self.id_col is not None:
                photons[self.id_col][intersect] = self.id_num
            # Set position in different coordinate systems
            photons['pos'][intersect] = interpos[intersect]
            photons[self.loc_coos_name[0]][intersect] = intercoos[intersect, 0]
            photons[self.loc_coos_name[1]][intersect] = intercoos[intersect, 1]

            if hasattr(self, "specific_process_photons"):
                outcols = self.specific_process_photons(photons, intersect, interpos, intercoos)
                self.add_output_cols(photons, list(outcols.keys()))
                for col in outcols:
                    if col == 'probability':
                        photons[col][intersect] *= outcols[col]
                    else:
                        photons[col][intersect] = outcols[col]

            elif hasattr(self, "process_photon"):
                if isinstance(photons, Row):
                    photons = Table(photons)
                outcols = ['dir', 'pos', 'energy', 'polarization', 'probability'] + self.output_columns
                n_intersect = intersect.nonzero()[0]
                for photon, i in zip(photons[intersect], n_intersect):
                    outs = self.process_photon(photon['dir'], photon['pos'],
                                               photon['energy'],
                                               photon['polarization'])
                    for a, b in zip(outcols, outs):
                        if a == 'probability':
                            photons['probability'][i] *= b
                        else:
                            photons[a][i] = b
            else:
                raise AttributeError('Optical element must have one of three: specific_process_photons, process_photon, or override process_photons.')

        return photons


class FlatOpticalElement(OpticalElement):
    '''Base class for geometrically flat optical elements.

    Compared with `OpticalElement` this adds methods to make the implementation of
    flat elements easier. It adds a default `geometry`, a fast, vectorized method `intersect`,
    and a template to `process_photons`.

    Derived classes have the option to implement their own `process_photons` or, alternatively,
    they can implement a function called
    ``specific_process_photons(self, photons, intersect, interpos, intercoos)`` that returns a dictionary
    of the form ``{'column name': value, ...}`` where value is an array that holds one value for
    each photon that intersects the optical element. In the special case of ``probability`` the
    return value should only contain the probability assigned in **this** element. This value
    will be multiplied with the previous probabilities of each photon automatically.
    '''

    display = {'shape': 'box'}
    _geometry = {}

    loc_coos_name = ['y', 'z']
    '''name for output columns that contain the interaction point in local coordinates.'''

    def __init__(self, *args, **kwargs):
        super(FlatOpticalElement, self).__init__(*args, **kwargs)
        #copy class attribute to instance attribute
        self._geometry = copy(self._geometry)

    def intersect(self, dir, pos):
        '''Calculate the intersection point between a ray and the element

        Parameters
        ----------
        dir : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of the direction of the ray
        pos : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of a point on the ray

        Returns
        -------
        intersect :  boolean array of length N
            ``True`` if an intersection point is found.
        interpos : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of the intersection point. Values are set
            to ``np.nan`` if no intersection point is found.
        interpos_local : `numpy.ndarray` of shape (N, 2)
            y and z coordinates in the coordiante system of the active plane
            (not normalized to the dimensions of the element in question, but
            in absolute units).
        '''
        plucker = dir_point2line(h2e(dir), h2e(pos))
        interpos =  intersect_line_plane(plucker, self.geometry('plane'))
        vec_center_inter = - h2e(self.geometry('center')) + h2e(interpos)
        ey = np.dot(vec_center_inter, h2e(self.geometry('e_y')))
        ez = np.dot(vec_center_inter, h2e(self.geometry('e_z')))
        intersect = ((np.abs(ey) <= np.linalg.norm(self.geometry('v_y'))) &
                     (np.abs(ez) <= np.linalg.norm(self.geometry('v_z'))))
        if dir.ndim == 2:
            interpos[~intersect, :3] = np.nan
        # input of single photon.
        elif (dir.ndim == 1) and not intersect:
            interpos[:3] = np.nan
        return intersect, interpos, np.vstack([ey, ez]).T


class FlatStack(FlatOpticalElement):
    '''Convenience class for several flat, stacked optical elements.

    This class is meant to simplify the specification of a single physical
    element, that fullfills several logical functions, e.g. a detector can be seen
    as a a sequence of a contamination layer (which modifies the probability of a photon
    reaching the CCD), a QE filter (which modifies the probability of detecting the photon),
    and the pixelated CCD (which sorts the photons in pixels). All these things can be
    approximated as happening in the same physical spotlocation, and thus it is convenient to
    treat all three functions as one element.

    Parameters
    ----------
    elements : list of classes
        List of class names specifying the layers in the stack
    keywords : list of dicts
        Dictionaries specifying the properties of each layer (do not set the position
        of individual elements)

    Examples
    --------
    In this example, we will define a single flat CCD with a QE of 0.5 for all energies.

    >>> from marxs.optics import FlatStack, EnergyFilter, FlatDetector
    >>> myccd = FlatStack(position=[0, 2, 2], zoom=2,
    ...     elements=[EnergyFilter, FlatDetector],
    ...     keywords=[{'filterfunc': lambda x: 0.5}, {'pixsize': 0.05}])

    '''

    def __init__(self, **kwargs):
        elements = kwargs.pop('elements')
        keywords = kwargs.pop('keywords')
        super(FlatStack, self).__init__(**kwargs)
        self.elements = []
        for elem, k in zip(elements, keywords):
            self.elements.append(elem(pos4d=self.pos4d, **k))

    def specific_process_photons(self, *args, **kwargs):
        return {}

    def process_photons(self, photons, intersect=None, interpos=None, intercoos=None):
        '''
        Parameters
        ----------
        intersect, interpos, intercoos : array (N, 4)
            The array ``interpos`` contains the intersection points in the global
            coordinate system, ``intercoos`` in the local (y,z) system of the grating.
        '''
        if intersect.sum() > 0:
            # This line calls FlatOpticalElement.process_photons to add ID cols and local coos
            # if requested (this could also be done by any of the contained sequence elements,
            # but we want the user to be able to specify that for either of them).
            photons = super(FlatStack, self).process_photons(photons, intersect, interpos, intercoos)
            for e in self.elements:
                photons = e.process_photons(photons, intersect, interpos, intercoos)

        return photons


def photonlocalcoords(f, colnames=['pos', 'dir']):
    '''Decorator for calculation that require a local coordinate system

    This is specifically meant to wrap the :meth:`process_photons` methods of
    any :class:`OpticalElement`; the current implementation expects the call
    signature of :meth:`process_photons`.

    This decorator transforms coordinates from the global system to the local
    system before a function call and back to the global system again after
    the function call.

    Parameters
    ----------
    f : callable with signature ``f(self, photons, *args, **kwargs)``
        The function to be decorated. In the signature, ``photons`` is a
        `astropy.table.Table`.
    colnames : list of strings
        List of all column names in the photon table to be transformed into a
        different coordinate system.
    '''
    @wraps(f)
    def wrapper(self, photons, *args, **kwargs):
        # transform to coordsys if single instrument
        invpos4d = np.linalg.inv(self.pos4d)
        for n in colnames:
            photons[n] = np.einsum('...ij,...j', invpos4d, photons[n])
        photons = f(self, photons, *args, **kwargs)
        # transform back into coordsys of satellite
        for n in colnames:
            photons[n] = np.einsum('...ij,...j', self.pos4d, photons[n])
        return photons

    return wrapper



import numpy as np
from matplotlib import pyplot as plt
from astropy.coordinates import SkyCoord

from marxs.source import poisson_process
from marxs.missions import chandra
from marxs.source import PointSource

ngc1313_X1 = SkyCoord("3h18m19.99s -66d29m10.97s")

energies = np.arange(.3, 8., .01)
spectrum1 = 6.3e-4 * energies**(-1.9)
flux1 = (spectrum1[1:] * np.diff(energies)).sum()
flux1 = poisson_process(flux1)

aperture = chandra.Aperture()

src1 = PointSource(coords=ngc1313_X1, energy={'energy': energies, 'flux': spectrum1},
                   flux=flux1, geomarea=aperture.area)

pointing = chandra.LissajousDither(coords=ngc1313_X1)
hrma = chandra.HRMA()
acis = chandra.ACIS(chips=[4,5,6,7,8,9], aimpoint=chandra.AIMPOINTS['ACIS-S'])

photons = src1.generate_photons(5e3)  # 5 ks exposure time
photons = pointing(photons)
photons = aperture(photons)
photons = hrma(photons)
photons = acis(photons)

line = plt.plot(photons['tdetx'], photons['tdety'], '.')



# Licensed under GPL version 3 - see LICENSE.rst
'''Add additional scattering to processed photons.

Classes in this file add additional scattering in a statistical sense.

The classes in this module do not trace the to a specific location, they just add
the scatter at the point of the last interaction.
For example, reflection from a (flat) mirror is implemented as a perfect reflection, but in
practice there is some roughness to the mirror that adds a small Gaussian blur to the
reflection. To represent this, the point of origin of the ray remains unchanged, but a small
random change is added to the direction vector.
'''
import numpy as np

from ..math.utils import e2h, h2e
from ..math.rotations import axangle2mat
from ..math.polarization import parallel_transport
from .base import FlatOpticalElement

class RadialMirrorScatter(FlatOpticalElement):
    '''Add scatter to any sort of radial mirror.

    Scatter is added in the plane of reflection, which is defined here
    as the plane which contains (i) the current direction the ray and (ii) the
    vector connecting the center of the `RadialMirrorScatter` element and the
    point of last interaction for the ray.
    Scatter can also be added perpendicular to the plane-of-reflection.

    Parameters
    ----------
    inplanescatter : float
        sigma of Gaussian for in-plane scatter [in radian]
    perpplanescatter : float
        sigma of Gaussian for scatter perpendicular to the plane of reflection
        [in radian] (default = 0)
    '''
    def __init__(self, **kwargs):
        self.inplanescatter = kwargs.pop('inplanescatter') # in rad
        self.perpplanescatter = kwargs.pop('perpplanescatter', 0.) # in rad
        super(RadialMirrorScatter, self).__init__(**kwargs)

    def specific_process_photons(self, photons, intersect, interpos, intercoos):
        # change this line, if you want to process only some photons (intersect et al.)
        n = intersect.sum()
        center = self.pos4d[:-1, -1]
        radial = h2e(photons['pos'][intersect].data) - center
        perpplane = np.cross(h2e(photons['dir'][intersect].data), radial)
        inplaneangle = np.random.normal(loc=0., scale=self.inplanescatter, size=n)

        rot = axangle2mat(perpplane, inplaneangle)
        outdir = e2h(np.einsum('...ij,...i->...j', rot, h2e(photons['dir'][intersect])), 0)

        if self.perpplanescatter !=0: # Works for 0 too, but waste of time to run
            perpangle = np.random.normal(loc=0., scale=self.perpplanescatter, size=n)
            rot = axangle2mat(radial, perpangle)
            outdir = e2h(np.einsum('...ij,...i->...j', rot, h2e(outdir)), 0)
        else:
            perpangle = np.zeros_like(inplaneangle)

        pol = parallel_transport(photons['dir'].data[intersect, :], outdir,
                                 photons['polarization'].data[intersect, :])

        return {'dir': outdir, 'polarization': pol,
                'inplanescatter': inplaneangle, 'perpplanescatter': perpangle}



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy import table
import astropy.units as u
from astropy.utils.metadata import enable_merge_strategies

from .base import FlatOpticalElement
from ..base import GeometryError
from ..visualization.utils import plane_with_hole
from ..math.utils import anglediff, h2e
from ..simulator import BaseContainer
from .. import utils


class BaseAperture(object):
    '''Base Aperture class'''

    display = {'color': (0.0, 0.75, 0.75),
               'opacity': 0.3,
               'shape': 'triangulation'}

    @staticmethod
    def add_colpos(photons):
        '''add columns ['pos'] to photon array'''
        photoncoords = table.Column(name='pos', length=len(photons), shape=(4,))
        photons.add_column(photoncoords)
        photons['pos'][:, 3] = 1

    @property
    def area(self):
        '''Area of the aperture.

        This does not take into account any projection effects for
        apertures that are not perpendicular to the optical axis.
        '''
        return NotImplementedError


class FlatAperture(BaseAperture, FlatOpticalElement):
    'Base class for geometrically flat apertures defined in python'
    def generate_local_xy(self, n):
        '''Generate x, y in the local coordinate system

        Parameters
        ----------
        n : int
            number of x, y coordinate pairs requested

        Returns
        -------
        x, y : arrays of floats
            (x, y) coordinates of the photons in the local frame
        '''
        return NotImplementedError

    def __call__(self, photons):
        self.add_output_cols(photons, self.loc_coos_name)
        # Add ID number to ID col, if requested
        if self.id_col is not None:
            photons[self.id_col] = self.id_num
        # Set position in different coordinate systems
        x, y = self.generate_local_xy(len(photons))
        if self.loc_coos_name is not None:
            photons[self.loc_coos_name[0]] = x
            photons[self.loc_coos_name[1]] = y
        photons['pos'] = self.geometry('center') + x.reshape((-1, 1)) * self.geometry('v_y') + y.reshape((-1, 1)) * self.geometry('v_z')
        projected_area = np.dot(photons['dir'].data, - self.geometry('e_x'))
        # Photons coming in "through the back" would have negative probabilities.
        # Unlikely to ever come up, but just in case we clip to 0.
        photons['probability'] *= np.clip(projected_area, 0, 1.)

        return photons

    def process_photons(self, photons):
        raise NotImplementedError('You probably want to use __call__.')

    def inner_shape(self):
        '''Return values in Eukledean space'''
        raise NotImplementedError

    def triangulate(self):
        '''Return a triangulation of the aperture hole embedded in a square.

        The size of the outer square is determined by the ``'outer_factor'`` element
        in ``self.display``.

        Returns
        -------
        xyz : np.array
            Numpy array of vertex positions in Eukeldian space
        triangles : np.array
            Array of index numbers that define triangles
        '''
        r_out = self.display.get('outer_factor', 3)
        g = self.geometry
        outer = h2e(g('center')) + r_out * np.vstack([h2e( g('v_y')) + h2e(g('v_z')),
                                                      h2e(-g('v_y')) + h2e(g('v_z')),
                                                      h2e(-g('v_y')) - h2e(g('v_z')),
                                                      h2e( g('v_y')) - h2e(g('v_z'))
        ])
        inner = self.inner_shape()
        return plane_with_hole(outer, inner)


class RectangleAperture(FlatAperture):
    '''Select the position where a parallel ray from an astrophysical source starts the simulation.

    '''
    def generate_local_xy(self, n):
        x = np.random.random(n) * 2. - 1.
        y = np.random.random(n) * 2. - 1.
        return x, y

    @property
    def area(self):
        '''Area covered by the aperture'''
        return 4 * np.linalg.norm(self.geometry('v_y')) * np.linalg.norm(self.geometry('v_z')) * u.mm**2

    def inner_shape(self):
        g = self.geometry
        return h2e(g('center')) + np.vstack([h2e( g('v_y')) + h2e(g('v_z')),
                                             h2e(-g('v_y')) + h2e(g('v_z')),
                                             h2e(-g('v_y')) - h2e(g('v_z')),
                                             h2e( g('v_y')) - h2e(g('v_z'))])


class CircleAperture(FlatAperture):
    '''Select the position where a parallel ray from an astrophysical source starts the simulation.

    Photons are placed in a circle. The radius of the circle is the lengths of its ``v_y`` vector.
    At this point, the aperture must have the same zoom in y and z direction; it cannot be used
    to simulate an entrance ellipse.

    Parameters
    ----------
    phi : list of 2 floats
        Bounding angles for a segment covered by the GSA. :math:`\phi=0`
        is on the positive y axis. The segment fills the space from ``phi1`` to
        ``phi2`` in the usual mathematical way (counterclockwise).
        Angles are given in radian. Use ``phi[1] < 0`` if the segment crosses
        the y axis. (Default is the full circle.)
    r_inner : float
        Inner radius for ring-like apertures. Default is 0 (full circle).
    '''
    def __init__(self, **kwargs):
        phi = kwargs.pop('phi', [0, 2. * np.pi])
        if np.max(np.abs(phi)) > 10:
            raise ValueError('Input angles >> 2 pi. Did you use degrees (radian expected)?')
        if phi[0] > phi[1]:
            raise ValueError('phi[1] must be greater than phi[0].')
        if (phi[1] - phi[0]) > (2 * np.pi + 1e-6):
            raise ValueError('phi[1] - phi[0] must be less than 2 pi.')
        self.phi = phi
        self.r_inner = kwargs.pop('r_inner', 0.)
        super(CircleAperture, self).__init__(**kwargs)
        if self.r_inner > np.linalg.norm(self.geometry('v_y')):
            raise ValueError('r_inner must be less than size of full aperture.')

        if not np.isclose(np.linalg.norm(self.geometry('v_y')),
                          np.linalg.norm(self.geometry('v_z'))):
            raise GeometryError('Aperture does not have the same size in y, z direction.')

    def generate_local_xy(self, n):
        phi = np.random.uniform(self.phi[0], self.phi[1], n)
        # normalize r_inner
        r_inner = self.r_inner / np.linalg.norm(self.geometry('v_y'))
        r = np.sqrt(np.random.uniform(r_inner**2, 1., n))

        x = r * np.cos(phi)
        y = r * np.sin(phi)
        return x, y

    @property
    def area(self):
        '''Area covered by the aperture'''
        A_circ = np.pi * (np.linalg.norm(self.geometry('v_y'))**2 - self.r_inner**2)
        return (self.phi[1] - self.phi[0])  / (2 * np.pi) * A_circ * u.mm**2

    def inner_shape(self, r_factor=1):
        n = self.display.get('n_inner_vertices', 90)
        phi = np.linspace(0.5 * np.pi, 2.5 * np.pi, n, endpoint=False)
        v_y = r_factor * self.geometry('v_y')
        v_z = r_factor * self.geometry('v_z')

        x = np.cos(phi)
        y = np.sin(phi)
        # phi could be less then full circle
        # wrap phi at lower bound (which could be negative).
        # For the default [0, 2 pi] this is a no-op
        phi = (phi - self.phi[0]) % (2 * np.pi)
        ind = phi < anglediff(self.phi)
        x[~ind] = 0
        y[~ind] = 0

        return h2e(self.geometry('center') + x.reshape((-1, 1)) * v_y + y.reshape((-1, 1)) * v_z)

    def triangulate(self):
        xyz, triangles = super(CircleAperture, self).triangulate()
        if (self.r_inner > 0):
            n_in = xyz.shape[0]
            n = self.display.get('n_inner_vertices', 90)
            new_xyz = self.inner_shape(r_factor = self.r_inner / np.linalg.norm(self.geometry('v_y')))
            xyz = np.vstack([xyz, h2e(self.geometry('center')), new_xyz])
            new_tri = np.zeros((n, 3))
            new_tri[:, 0] = n_in
            new_tri[:, 1] = np.arange(n_in + 1, n_in + n + 1)
            new_tri[0, 2] = n_in + n
            new_tri[1:, 2] = np.arange(n_in + 1, n_in + n)
            triangles = np.vstack([triangles, new_tri])
        return xyz, triangles


class MultiAperture(BaseAperture, BaseContainer):
    '''Group several apertures into one class.

    Sometimes a single intrument has several physical openings where photons from an
    astrophysical source can enter, an example is XMM-Newton that operates three telescopes
    in parallel. While it is often more efficient to simulate these as entirely separate by
    running separate simulations, that is not always true.
    This class groups several apertures together.

    .. warning::

       Apertures cannot overlap. There is currently no code checking for this, but
       overlapping apertures will produce unphysical results.


    Parameters
    ----------
    elements : list
        The elements of this list are all optical elements that process photons.
    preprocess_steps : list
        The elements of this list are functions or callable objects that accept a photon list as input
        and return no output (*default*: ``[]``). All ``preprocess_steps`` are run before
        *every* aperture on just the photons that pass this aperture.
    postprocess_steps : list
        See ``preprocess_steps`` except that the steps are run *after* each aperture
         (*default*: ``[]``) on just the photons that passed that aperture.
    '''
    display = {'shape': 'container'}

    def __init__(self, **kwargs):
        self.elements = kwargs.pop('elements')
        self.id_col = kwargs.pop('id_col', 'aperture')
        super(MultiAperture, self).__init__(**kwargs)

    @property
    def area(self):
        '''Area covered by the aperture'''
        return u.Quantity([e.area for e in self.elements]).sum()

    def __call__(self, photons):
        areas = u.Quantity([e.area for e in self.elements])
        aperid = np.digitize(np.random.rand(len(photons)), np.cumsum(areas) / self.area)

        # Add ID number to ID col, if requested
        if self.id_col is not None:
            photons[self.id_col] = aperid
        outs = []
        for i, elem in enumerate(self.elements):
            thisphot = photons[aperid == i]
            for p in self.preprocess_steps:
                p(thisphot)
            thisphot = elem(thisphot)
            for p in self.postprocess_steps:
                p(thisphot)
            outs.append(thisphot)
        with enable_merge_strategies(utils.MergeIdentical):
            photons = table.vstack(outs)

        return photons



# Licensed under GPL version 3 - see LICENSE.rst
'''This module collects helper functions for visualization that are of use for several backends.

The functions here are not intended to be called directly by the user. Instead, they
refractor common tasks that are used in several visualization backends.
'''
from __future__ import division

import warnings
import numpy as np


class MARXSVisualizationWarning(Warning):
    '''Warning class for MARXS objects missing from plotting'''
    pass


def get_obj_name(obj):
    '''Return printable name for objects or functions.'''
    if hasattr(obj, 'name'):
        return obj.name
    elif hasattr(obj, 'func_name'):
        return obj.func_name
    else:
        return str(obj)


def plot_object_general(plot_registry, obj, display=None, **kwargs):
    '''Look up a plottig routine for an object and execute it.

    This function is not meant to be called directly by the user, instead, it
    is designed to simplify the implementation of new plotting backends.

    Parameters
    ----------
    plot_registry : dict
        Keys are the names of the shape of an object and values in this
        dictionary are functions that know how to plot this type of shape. The
        appropriate plotting function is then called with the input `obj`,
        `display` and any other keyword arguments.
        If the shape is ``"None"`` (as a string), no plotting function is
        called.
    obj : `marxs.base.MarxsElement`
        The element that should be plotted.
    display : dict of None
        Dictionary with display settings. If this is ``None``, ``obj.display``
        is used. If that is also ``None`` then the objects is skipped.
    kwargs : other keyword arguments
        These arguments are just passed through to the plotting function.

    Returns
    -------
    out : different
        The output from the plotting function that was executed is passed
        through.  Different plotting backends return different kinds of output.
    '''
    if display is None:
        if hasattr(obj, 'display') and (obj.display is not None):
            display = obj.display
        else:
            warnings.warn('Skipping {0}: No display dictionary found.'.format(get_obj_name(obj)),
                           MARXSVisualizationWarning)
            return None

    out = None
    if 'shape' in display:
        shape = display['shape']
        shapes = [s.strip() for s in shape.split(';')]
        for s in shapes:
            if s == 'None':
                break
            elif s in plot_registry:
                # turn into valid color tuple
                display['color'] = get_color(display)
                out = plot_registry[s](obj, display,  **kwargs)
                break
        else:
            warnings.warn('Skipping {0}: No function to plot {1}.'.format(get_obj_name(obj), shape),
                          MARXSVisualizationWarning)
    else:
        warnings.warn('Skipping {0}: "shape" not set in display dict.'.format(get_obj_name(obj)),
                       MARXSVisualizationWarning)
    return out


def get_color(d):
    '''Look for color information in dictionary.

    If missing, return white.
    This function checks if the `d['color']` is a valid RGB tuple and if
    not it imports a `matplotlib.colors.ColorConverter` to convert any
    matplotlib compatible string to an RGB tuple.

    Parameters
    ----------
    d : dict
        Color information should be present in ``d['color']``.

    Returns
    -------
    color : tuple
        RGB tuple with each element in the range 0..1
    '''
    if 'color' not in d:
        return (1., 1., 1.)
    else:
        c = d['color']
        # check if this is a tuple of three floats n the range 0..1
        # or can be converted to one
        try:
            cout = tuple(c)
            if len(cout) != 3:
                raise TypeError
            for a in cout:
                if not isinstance(a, float) or (a < 0.) or (a > 1.):
                    raise TypeError
            return cout
        except TypeError:
            # It's a hex or string. Let matplotlib deal with that.
            import matplotlib.colors
            return matplotlib.colors.colorConverter.to_rgb(c)


def color_tuple_to_hex(color):
    '''Convert color tuple to hex string.

    Parameters
    ----------
    color : tuple
        tuple has three elements (rgb) that are floats betwen 0 and 1
        or ints between 0 and 255.

    Returns
    -------
    hexstring : string
        string encoding that number as hex
    '''
    if all([isinstance(a, float) for a in color]):
        if any(i < 0. for i in color) or any(i > 1. for i in color):
            raise ValueError('Float values in color tuple must be between 0 and 1.')
        out = hex(int(color[0] * 256**2 * 255 + color[1] * 256 * 255 + color[2] * 255))
    elif all([isinstance(a, int) for a in color]):
        if any(i < 0 for i in color) or any(i > 255 for i in color):
            raise ValueError('Int values in color tuple must be between 0 and 255.')
        out = hex(color[0] * 256**2 + color[1] * 256 + color[2])
    else:
        raise ValueError('Input tuple must be all float or all int.')
    # Now pad with zeros if required
    return out[:2] + out[2:].zfill(6)


def plane_with_hole(outer, inner):
    '''Triangulation of a plane with an inner hole

    This function constructs a triangulation for a plane with an inner hole, e.g.
    a rectangular plane where an inner circle is cut out.

    Parameters
    ----------
    outer, inner : np.ndarray of shape (n, 3)
        Coordinates in x,y,z of points that define the inner and outer
        boundary.  ``outer`` and ``inner`` can have a different number of
        points, but points need to be listed in the same orientation
        (e.g. clockwise) for both and the starting points need to have a
        similar angle as seen from the center (e.g. for a plane with z=0, both
        ``outer`` and ``inner`` could list a point close to the y-axis first.

    Returns
    -------
    xyz : nd.array
        stacked ``outer`` and ``inner``.
    triangles : nd.array
        List of the indices. Each row has the index of three points in ``xyz``.

    Examples
    --------
    In this example, we make a square and cut out a smaller square in the middle.

    >>> import numpy as np
    >>> from marxs.visualization.utils import plane_with_hole
    >>> outer = np.array([[-1, -1, 1, 1], [-1, 1, 1, -1], [0,0,0,0]]).T
    >>> inner = 0.5 * outer
    >>> xyz, triangles = plane_with_hole(outer, inner)
    >>> triangles
    array([[0, 4, 5],
       [0, 1, 5],
       [1, 5, 6],
       [1, 2, 6],
       [2, 6, 7],
       [2, 3, 7],
       [3, 7, 4],
       [3, 0, 4]])

    '''
    n_out = outer.shape[0]
    n_in = inner.shape[0]
    n = n_out + n_in

    triangles = np.zeros((n, 3), dtype=int)
    xyz = np.vstack([outer, inner])

    i_in = 0
    i_out = 0
    for i in range(n_out + n_in):
        if i/n >= i_in/n_in:
            triangles[i, :] = [i_out, n_out + i_in, n_out + ((i_in + 1) % n_in)]
            i_in += 1
        else:
            triangles[i, :] = [i_out, (i_out + 1) % n_out, n_out + (i_in % n_in)]
            i_out = (i_out + 1) % n_out
    return xyz, triangles



# Licensed under GPL version 3 - see LICENSE.rst
from __future__ import division
import warnings

import numpy as np
from transforms3d.affines import decompose44, compose

from .base import FlatOpticalElement, OpticalElement
from ..math.utils import h2e
from ..visualization.utils import get_color
from ..utils import SimulationSetupWarning


class FlatDetector(FlatOpticalElement):
    '''Flat detector with square pixels

    Processing the a photon with this detector adds four columns to the photon
    properties:

    - ``det_x``, ``det_y``: Event position on the detector in mm
    - ``detpix_x``, ``detpix_y``: Event position in detector coordinates, where
      (0, 0) is on the corner of the chip.

    The (x,y) coordinate system on the chip is such that it falls on the (y,z) plane
    of the global x,y,z coordinate system (it would be more logical to call the chip
    system (y,z), but traditionally that is not how chip coordinates are named). The
    pixel in the corner has coordinates (0, 0) in the pixel center.

    Parameters
    ----------
    pixsize : float
        size of pixels in mm

    kwargs :
       see `args for optical elements`
    '''

    loc_coos_name = ['det_x', 'det_y']
    '''name for output columns that contain the interaction point in local coordinates.'''

    detpix_name = ['detpix_x', 'detpix_y']
    '''name for output columns that contain this pixel number.'''

    display = {'color': (1.0, 1.0, 0.),
               'shape': 'box'}

    def __init__(self, pixsize=1, **kwargs):
        self.pixsize = pixsize
        super(FlatDetector, self).__init__(**kwargs)
        t, r, zoom, s = decompose44(self.pos4d)
        self.npix = [0, 0]
        self.centerpix = [0, 0]
        for i in (0, 1):
            z  = zoom[i + 1]
            self.npix[i] = int(np.round(2. * z / self.pixsize))
            if np.abs(2. * z / self.pixsize - self.npix[i]) > 1e-2:
                warnings.warn('Detector size is not an integer multiple of pixel size in direction {0}. It will be rounded.'.format('xy'[i]), SimulationSetupWarning)
            self.centerpix[i] = (self.npix[i] - 1) / 2

    def specific_process_photons(self, photons, intersect, interpos, intercoos):
        detx = intercoos[intersect, 0] / self.pixsize + self.centerpix[0]
        dety = intercoos[intersect, 1] / self.pixsize + self.centerpix[1]
        return {self.detpix_name[0]: detx, self.detpix_name[1]: dety}

class CircularDetector(OpticalElement):
    '''A detector shaped like a ring or tube.

    This detector is shaped like a tube. The form is a circle in the xy plane and
    and flat along the z-direction.
    While most CCDs are flat in practice, the `CircularDetector` simulates a setup
    that can follow the Rowland circle geometry exactly which is useful, e.g. to study
    the resolution of a spectrograph without worrying about the details of the detector
    geometry.

    Parameters
    ----------
    position, orientation, zoom, pos4d : see description of `pos4d`
        The radius of the tube is given by the ``zoom`` keyword, see `pos4d`.
        Use ``zoom[0] == zoom[1]`` to make a circular tube. ``zoom[0] != zoom[1]`` gives
        an elliptical profile. ``zoom[2]`` sets the extension in the z direction.
    pixsize : float
        size of pixels in mm
    phi_offset : float
        This defines the center (pixel = 0) in radian. In other words:
        The output ``det_phi`` column has its zero position at ``phi_offset`` and
        values for ``phi`` are in the range [-pi, pi].
    '''
    loc_coos_name = ['det_phi', 'det_y']

    detpix_name = ['detpix_x', 'detpix_y']
    '''name for output columns that contain this pixel number.'''

    display = {'color': (1.0, 1.0, 0.),
               'opacity': 0.7,
               'shape': 'surface',
               'coo1': np.linspace(0, 2 * np.pi, 50),
               'coo2': [-1, 1]}

    def __init__(self, pixsize=1, **kwargs):
        self.pixsize = pixsize
        self.phi_offset = kwargs.pop('phi_offset', 0.)
        self.inwards = kwargs.pop('inside', True)
        super(CircularDetector, self).__init__(**kwargs)

    @classmethod
    def from_rowland(cls, rowland, width):
        '''Generate a `CircularDetector` from a `RowlandTorus`.

        Parameters
        ----------
        rowland : `~marxs.design.RowlandTorus`
            The circular detector is constructed to fit exactly into the
            Rowland Circle defined by ``rowland``.
        width : float
            Half-width of the tube in the flat direction (z-axis) in mm
        '''
        # Step 1: Get position and size from Rowland torus
        pos4d_circ = compose([rowland.R, 0, 0], np.eye(3), [rowland.r, rowland.r, width])
        # Step 2: Transform to global coordinate system
        pos4d_circ = np.dot(rowland.pos4d, pos4d_circ)
        # Step 3: Make detector
        return cls(pos4d=pos4d_circ, phi_offset=-np.pi)

    @property
    def _inwardsoutwards(self):
        'Transform the self.inwards bool into [-1, +1]'
        if self.inwards:
            return 1.
        else:
            return -1.

    def intersect(self, dir, pos, transform=True):
        '''Calculate the intersection point between a ray and the element

        Parameters
        ----------
        dir : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of the direction of the ray
        pos : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of a point on the ray
        transform : bool
            If ``True``, input is in global coordinates and needs to be transformed
            here for the calculations; if ``False`` input is in local coordinates.

        Returns
        -------
        intersect :  boolean array of length N
            ``True`` if an intersection point is found.
        interpos : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of the intersection point. Values are set
            to ``np.nan`` is no intersecton point is found.
        interpos_local : `numpy.ndarray` of shape (N, 2)
            phi, z coordiantes (in the local frame) for one of the intersection points.
            If both intersection points are required, reset ``self.inner`` and call this
            function again.
        '''
        # This could be moved to a general function
        if not np.all(dir[:, 3] == 0):
            raise ValueError('First input must be direction vectors.')
        # Could test pos, too...
        if transform:
            invpos4d = np.linalg.inv(self.pos4d)
            dir = np.dot(invpos4d, dir.T).T
            pos = np.dot(invpos4d, pos.T).T

        xyz = h2e(pos)

        # Solve quadratic equation in steps. a12 = (-xr +- sqrt(xr - r**2(x**2 - R**2)))
        xy = xyz[:, :2]
        r = dir[:, :2]
        underroot = (np.einsum('ij,ij->i', xy, r))**2 - np.sum(r**2, axis=1) * (np.sum(xy**2, axis=1) - 1.)
        intersect = (underroot >= 0)
        i = intersect  # just a shorthand because it's used so much below

        interpos_local = np.ones((pos.shape[0], 2))
        interpos_local[:] = np.nan
        interpos = np.ones_like(pos)
        interpos[:] = np.nan

        if intersect.sum() > 0:
            b = np.sum(xy[i] * r[i], axis=1)
            denom = np.sum(r[i]**2, axis=1)
            a1 = (- b + np.sqrt(underroot[i])) / denom
            a2 = (- b - np.sqrt(underroot[i])) / denom
            x1 = xy[i, :] + a1[:, np.newaxis] * r[i, :]
            apick = np.where(self._inwardsoutwards * np.sum(x1 * r[i, :], axis=1) >=0, a1, a2)
            xy_p = xy[i, :] + apick[:, np.newaxis] * r[i, :]
            phi = np.arctan2(xy_p[:, 1], xy_p[:, 0])
            # Shift phi by offset, then wrap to that it is in range [-pi, pi]
            interpos_local[i, 0] = (phi - self.phi_offset + np.pi) % (2 * np.pi) - np.pi
            # Those look like they hit in the xy plane.
            # Still possible to miss if z axis is too large.
            # Calculate z-coordiante at intersection
            interpos_local[intersect, 1] = xyz[i, 2] + apick * dir[i, 2]
            interpos[i, :2] = xy_p
            interpos[i, 2] = interpos_local[i, 1]
            interpos[i, 3] = 1
            # set those elements on intersect that miss in z to False
            trans, rot, zoom, shear = decompose44(self.pos4d)
            z_p = interpos[i, 2]
            intersect[i.nonzero()[0][np.abs(z_p) > 1]] = False
            # Now reset everything to nan that does not intersect
            interpos_local[~i, :] = np.nan
            # interpos_local in z direction is in local coordinates, i.e.
            # the x coordiante is 0..1, but we want that in units of the
            # global coordinate system.
            interpos_local[:, 1] = interpos_local[:, 1] * zoom[2]
            interpos[~i, :] = np.nan

            interpos = np.dot(self.pos4d, interpos.T).T

        return intersect, interpos, interpos_local

    def process_photons(self, photons, intersect, interpos, inter_local):
        photons['pos'][intersect, :] = interpos[intersect, :]
        self.add_output_cols(photons, self.loc_coos_name + self.detpix_name)
        # Add ID number to ID col, if requested
        if self.id_col is not None:
            photons[self.id_col][intersect] = self.id_num
        # Set position in different coordinate systems
        photons['pos'][intersect] = interpos[intersect]
        photons[self.loc_coos_name[0]][intersect] = inter_local[intersect, 0]
        photons[self.loc_coos_name[1]][intersect] = inter_local[intersect, 1]
        trans, rot, zoom, shear = decompose44(self.pos4d)
        if np.isclose(zoom[0], zoom[1]):
            photons[self.detpix_name[0]][intersect] = inter_local[intersect, 0] * zoom[0] / self.pixsize
        else:
            warnings.warn('Pixel coordinate for elliptical mirrors not implemented.', PixelSizeWarning)
        photons[self.detpix_name[1]][intersect] = inter_local[intersect, 1] / self.pixsize

        return photons

    def parametric_surface(self, phi, z=np.array([-1, 1])):
        '''Parametric description of the tube.

        This is just another way to obtain the shape of the tube, e.g.
        for visualization.

        Parameters
        ----------
        phi : np.array
            ``phi`` is the angle around the tube profile.
        z : np.array
            The coordiantes along the radius coordinate.

        Returns
        -------
        xyzw : np.array
            Ring coordinates in global homogeneous coordinate system.
        '''
        phi = np.asanyarray(phi)
        z = np.asanyarray(z)
        if (phi.ndim != 1) or (z.ndim != 1):
            raise ValueError('input parameters have 1-dim shape.')
        phi, z = np.meshgrid(phi, z)
        x = np.cos(phi)
        y = np.sin(phi)
        w = np.ones_like(z)
        coos = np.array([x, y, z, w]).T
        return np.einsum('...ij,...j', self.pos4d, coos)



# Licensed under GPL version 3 - see LICENSE.rst
from __future__ import division

import numpy as np
import scipy.optimize
from astropy.stats import sigma_clipped_stats

from ..optics import FlatDetector

def sigma_clipped_std(data, **kwargs):
    '''Return stddev of sigma-clipped data.

    Parameters
    ----------
    data : np.array
        unbinned input data
    kwargs : see `astropy.stats.sigma_clipped_stats`
        All keyword arguments are passed to `astropy.stats.sigma_clipped_stats`.
    '''
    mean, median, std = sigma_clipped_stats(data, **kwargs)
    return std


def find_best_detector_position(photons, col='det_x',
                                objective_func=sigma_clipped_std,
                                orientation=np.eye(3), **kwargs):
    '''Numerically find the position of best focus.

    This routine places detectors at different positions and
    calculates the width of the photon distribution for each position.
    As a side effect, ``photons`` will be set to the intersection with
    the last try of the detector position.

    Parameters
    ----------
    photons : `astropy.table.Table`
        Input photon list.
    col : string
        Column name of the photon distribution to be minimized.
        The default is set for detectors that look for a grating signal
        (which is dispersed in ``det_y`` direction).
    objective_func : function
        Function that accepts a np.array as input and return the width.
    rotation : np.array of shape (3,3)
        Rotation matrix for the detector. By default the detector is parallel
        to the yz plane of the global coordinate system (see `pos4d`).
    kwargs : see `scipy.optimize.minimize`
        All other keyword argument will be passed to `scipy.optimize.minimize`.
    Returns
    -------
    opt : OptimizeResult
        see `scipy.optimize.minimize`

    '''
    def width(x, photons):
        mdet = FlatDetector(position=np.array([x, 0, 0]), orientation=orientation, zoom=1e5, pixsize=1.)
        photons = mdet(photons)
        return objective_func(photons[col].data)

    return scipy.optimize.minimize(width, 0, args=(photons,), options={'maxiter': 20, 'disp': True},
                                   **kwargs)


def detected_fraction(photons, labels, col='order'):
    '''Calculate the fraction of photons detected for some integer label

    While written for calculating Aeff per order, this can be used with any discrete
    quantity, e.g. Aeff per CCD.

    Parameters
    ----------
    photons : `astropy.table.Table`
        Photon event list
    labels : np.array
        Numeric (integer) labels that are found in column ``col``. When, e.g.,
        the effective area per order is calculated, this array should contain
        the order numbers for which the calculation will be done.
    col : string
        Column name for the order column.

    Returns
    -------
    prop : np.array
        Probability for a photon in a specific order to be detected.
    '''
    prob = np.zeros_like(labels, dtype=float)
    for i, o in enumerate(labels):
        ind = (photons[col] == o)
        if filterfunc is not None:
            ind = ind & filterfunc(photons)
        prob[i] = np.sum(photons['probability'][ind]) / len(photons)
    return prob



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy.table import Table
import transforms3d

from ..detector import FlatDetector, CircularDetector
from ...tests import closeornan
from ...math.utils import h2e
from ...design import RowlandTorus

def test_pixelnumbers():
    pos = np.array([[0, 0., -0.25, 1.],
                    [0., 9.9, 1., 1.],
                    [0., 10.1, 1., 1.]])
    dir = np.ones((3,4), dtype=float)
    dir[:, 3] = 0.
    photons = Table({'pos': pos, 'dir': dir,
                     'energy': [1,2., 3.], 'polarization': [1.,2, 3.], 'probability': [1., 1., 1.]})
    det = FlatDetector(zoom=[1., 10., 5.], pixsize=0.5)
    assert det.npix == [40, 20]
    assert det.centerpix == [19.5, 9.5]
    photons = det(photons)
    assert closeornan(photons['det_x'], np.array([0, 9.9, np.nan]))
    assert closeornan(photons['det_y'], np.array([-0.25, 1., np.nan]))
    assert closeornan(photons['detpix_x'], np.array([19.5, 39.3, np.nan]))
    assert closeornan(photons['detpix_y'], np.array([9, 11.5, np.nan]))

    # Regression test: This was rounded down to [39, 39] at some point.
    det = FlatDetector(pixsize=0.05)
    assert det.npix == [40, 40]


def test_nonintegerwarning(recwarn):
    det = FlatDetector(zoom=np.array([1.,2.,3.]), pixsize=0.3)
    w = recwarn.pop()
    assert 'is not an integer multiple' in str(w.message)

def test_intersect_tube_miss():
    '''Rays passing at larger radius or through the center miss the tube.'''
    circ = CircularDetector()
    # Passing at larger radius
    intersect, interpos, inter_local = circ.intersect(np.array([[1., 0., .1, 0],
                                                                [-1., -1., 0., 0.]]),
                                                         np.array([[0, -1.1, 0., 1.],
                                                                   [2., 0., 0., 1.]]))
    assert np.all(intersect == False)
    assert np.all(np.isnan(interpos))
    assert np.all(np.isnan(inter_local))

    # Through the center almost parallel to z-axis
    intersect, interpos, inter_local = circ.intersect(np.array([[0., 0.1, 1., 0]]),
                                                         np.array([[0.5, 0.5, 0., 1.]]))
    assert np.all(intersect == False)
    assert np.all(np.isnan(interpos))
    assert np.all(np.isnan(inter_local))

    # Repeat with a tube that's moved to make sure we did not mess up local and
    # global coordinates in the implementation
    circ = CircularDetector(position=[0.8, 0.8, 1.2])
    intersect, interpos, inter_local = circ.intersect(np.array([[1., 0., .0, 0],
                                                                [1., 0., 0., 0.]]),
                                                         np.array([[0., 0., 0., 1.],
                                                                   [0., -0.3, 1., 1.]]))
    assert np.all(intersect == False)
    assert np.all(np.isnan(interpos))
    assert np.all(np.isnan(inter_local))

def test_intersect_tube_hitmiss_zoomed():
    '''Rays that hit a one tube, but miss a shorter one'''
    dir = np.array([[-.1, 0., 0., 0], [1., 0, -0.2, 0.]])
    pos = np.array([[20., 0., 0.8, 1.], [2., .0, 0.4, 1.]])

    circ = CircularDetector(zoom=[1, 1, 1])
    intersect, interpos, inter_local = circ.intersect(dir, pos)
    assert np.all(intersect == True)

    circ = CircularDetector(zoom=[1, 1, .5])
    intersect, interpos, inter_local = circ.intersect(dir, pos)
    assert np.all(intersect == False)
    assert np.all(np.isnan(interpos))
    assert np.all(np.isnan(inter_local))

def test_intersect_tube_2points():
    dir = np.array([[-.1, 0., 0., 0], [-0.5, 0, 0., 0.], [-1., 0., 0., 0.]])
    pos = np.array([[50., 0., 0., 1.], [10., .5, 0., 1.], [2, 0, .3, 1.]])

    circ = CircularDetector(phi_offset=-np.pi)
    intersect, interpos, inter_local = circ.intersect(dir, pos)
    assert np.all(intersect == True)
    assert np.allclose(h2e(interpos), np.array([[-1., 0., 0.],
                                                [-np.sqrt(0.75), 0.5, 0.],
                                                [-1, 0., .3]]))
    assert np.allclose(inter_local, np.array([[0., -np.arcsin(0.5), 0.],
                                              [0., 0., 0.3]]).T)

def test_intersect_tube_2points_zoomed():
    '''inter_local should be the same if we make the tube longer in z'''
    dir = np.array([[-.1, 0., 0., 0], [-0.5, 0, 0., 0.], [-1., 0., 0., 0.]])
    pos = np.array([[50., 0., 0., 1.], [10., .5, 0., 1.], [2, 0, .3, 1.]])

    circ = CircularDetector(phi_offset=-np.pi, zoom=[1, 1., 5.])
    intersect, interpos, inter_local = circ.intersect(dir, pos)
    assert np.all(intersect == True)
    assert np.allclose(h2e(interpos), np.array([[-1., 0., 0.],
                                                [-np.sqrt(0.75), 0.5, 0.],
                                                [-1, 0., .3]]))
    assert np.allclose(inter_local, np.array([[0., -np.arcsin(0.5), 0.],
                                              [0., 0., 0.3]]).T)

def test_intersect_tube_2points_outside():
    '''Now hit the outside. Looks very similar, if we remove the phi_offset, too.'''
    dir = np.array([[-.1, 0., 0., 0], [-0.5, 0, 0., 0.], [-1., 0., 0., 0.]])
    pos = np.array([[50., 0., 0., 1.], [10., .5, 0., 1.], [2, 0, .3, 1.]])

    circ = CircularDetector(inside=False)
    intersect, interpos, inter_local = circ.intersect(dir, pos)
    assert np.all(intersect == True)
    assert np.allclose(h2e(interpos), np.array([[1., 0., 0.],
                                                [np.sqrt(0.75), 0.5, 0.],
                                                [1, 0., .3]]))
    assert np.allclose(inter_local, np.array([[0., np.arcsin(0.5), 0.],
                                              [0., 0., 0.3]]).T)

def test_intersect_tube_2points_translation():
    '''Repeat with a tube that's moved and zoomed to make sure we
    did not mess up local and global coordinates in the implementation
    '''
    circ = CircularDetector(position=[0.8, 0.8, 1.2], zoom=[1, 1, 2])
    intersect, interpos, inter_local = circ.intersect(np.array([[1., 0., .0, 0],
                                                                [1., 0., 0., 0.]]),
                                                         np.array([[0., 0., 0., 1.],
                                                                   [0., 0., 1., 1.]]))
    assert np.all(intersect == True)
    assert np.allclose(h2e(interpos), np.array([[1.4, 0., 0.],
                                                [1.4, 0., 1.]]))

def test_tube_parametric():
    '''Generate points on surface using parametic. Then make rays for intersection
    which should return those points as intersection points.'''
    circ = CircularDetector(zoom=[2., 3., 4.])
    parametric = circ.parametric_surface(phi=[-.1, 0., 1.])
    parametric = parametric.reshape((6, 4))
    # select dir so that it's in the right direction to recover these points.
    dir = np.tile([1, 0, 0, 0], (6, 1))
    intersect, interpos, inter_local = circ.intersect(dir, parametric)
    assert np.allclose(parametric, interpos)

def test_CircularDetector_from_Rowland():
    '''If a circular detector is very narrow, then all points should be
    very close to the Rowland torus.'''
    rowland = RowlandTorus(R=6e4, r=5e4, position=[123., 345., -678.],
                           orientation=transforms3d.euler.euler2mat(1, 2, 3, 'syxz'))
    detcirc = CircularDetector.from_rowland(rowland, width=1e-6)
    phi = np.mgrid[0:2.*np.pi:360j]
    points = detcirc.parametric_surface(phi)
    # Quartic < 1e5 is very close for these large values of r and R.
    assert np.max(np.abs(rowland.quartic(h2e(points)))) < 1e5



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import pytest
from scipy.stats import normaltest
from astropy.table import Table
from astropy.coordinates import SkyCoord
import astropy.units as u

from ...source import (SymbolFSource, Source, poisson_process,
                       PointSource,
                       DiskSource, SphericalDiskSource, GaussSource)
from ...optics import RectangleAperture
from ..source import SourceSpecificationError


def test_photons_header():
    '''All generated photons should have some common keywords.

    Just test that some of the keywords are there. It's unlikely
    that I need a full list here.
    '''
    s = SymbolFSource(coords=SkyCoord(-123., -43., unit=u.deg), size=1.*u.deg)
    photons = s.generate_photons(5.)
    for n in ['EXPOSURE', 'CREATOR', 'MARXSVER', 'SIMTIME', 'SIMUSER']:
        assert n in photons.meta

def test_energy_input_default():
    '''For convenience and testing, defaults for time, energy and pol are set.'''
    s = Source()
    photons = s.generate_photons(5.)
    assert len(photons) == 5
    assert np.all(photons['energy'] == 1.)
    assert len(set(photons['polangle'])) == 5  # all pol angles are different

def test_flux_input():
    '''Options: contant rate or function'''
    # 1. constant rate
    s = Source(flux=5)
    photons = s.generate_photons(5.)
    assert len(photons) == 25
    delta_t = np.diff(photons['time'])
    assert np.allclose(delta_t, delta_t[0]) # constante rate

    # 2. function
    def f(t, a):
        return np.logspace(1, np.log10(t))

    s = Source(flux=f)
    photons = s.generate_photons(100)
    assert np.all(photons['time'] == np.logspace(1, 2))

    # 3. anything else
    s = Source(flux=ValueError())
    with pytest.raises(SourceSpecificationError) as e:
        photons = s.generate_photons(5)
    assert '`flux` must be' in str(e.value)

def test_energy_input():
    '''Many different options ...'''
    # 1. contant energy
    s = PointSource(coords=SkyCoord("1h12m43.2s +1d12m43s"), energy=2.)
    photons = s.generate_photons(5)
    assert np.all(photons['energy'] == 2.)

    # 2. function
    def f(t):
        return t
    s = Source(energy=f)
    photons = s.generate_photons(5)
    assert np.all(photons['energy'] == photons['time'])
    # bad function
    s = Source(energy=lambda x: np.array([np.sum(x)]))
    with pytest.raises(SourceSpecificationError) as e:
        photons = s.generate_photons(5)
    assert 'an array of same size as' in str(e.value)

    # 3. array, table, recarray, dict, ... just try some of of the opion with keys
    # spectrum with distinct lines
    engrid = [0.5, 1., 2., 3.]
    fluxgrid = [456., 1., 0., 2.]  # first entry (456) will be ignored
    s1 = Source(energy={'energy': engrid, 'flux': fluxgrid})
    s2 = Source(energy=np.vstack([engrid, fluxgrid]))
    s3 = Source(energy=Table({'energy': engrid, 'flux': fluxgrid}))
    for s in [s1, s2, s3]:
        photons = s.generate_photons(1000)
        en = photons['energy']
        ind0510 = (en >= 0.5) & (en <=1.0)
        ind2030 = (en >= 2.) & (en <=3.0)
        assert (ind0510.sum() + ind2030.sum()) == len(photons)
        assert ind0510.sum() < ind2030.sum()

    # 4. anything else
    s = Source(energy=object())
    with pytest.raises(SourceSpecificationError) as e:
        photons = s.generate_photons(5)
    assert '`energy` must be' in str(e.value)

def test_polarization_input():
    '''Many differnet options ...'''
    # 1. 100 % polarized flux
    s = Source(polarization=2.)
    photons = s.generate_photons(5)
    assert np.all(photons['polangle'] == 2.)

    # 2. function
    def f(t, en):
        return t * en
    s = Source(polarization=f, energy=2.)
    photons = s.generate_photons(5)
    assert np.all(photons['polangle'] == photons['time'] * photons['energy'])
    # bad function
    s = Source(polarization=lambda x, y: np.array([np.dot(x, y)]))
    with pytest.raises(SourceSpecificationError) as e:
        photons = s.generate_photons(5)
    assert 'an array of same size as' in str(e.value)

    # 3. array, table, recarray, dict, ... just try some of of the opion with keys
    # spectrum with distinct lines
    polgrid = [0.5, 1., 2., 3.]
    probgrid = [456., 1., 0., 2.]  # first entry (456) will be ignored
    s1 = Source(polarization={'angle': polgrid, 'probability': probgrid})
    s2 = Source(polarization=np.vstack([polgrid, probgrid]))
    s3 = Source(polarization=Table({'angle': polgrid, 'probability': probgrid}))
    for s in [s1, s2, s3]:
        photons = s.generate_photons(1000)
        pol = photons['polangle']
        ind0510 = (pol >= 0.5) & (pol <=1.0)
        ind2030 = (pol >= 2.) & (pol <=3.0)
        assert (ind0510.sum() + ind2030.sum()) == len(photons)
        assert ind0510.sum() < ind2030.sum()

    # 4. None (unpolarized source)
    s = Source(polarization=None)
    photons = s.generate_photons(5)
    assert len(set(photons['polangle'])) == len(photons) # all different

    # 5. anything else
    s = Source(polarization=object())
    with pytest.raises(SourceSpecificationError) as e:
        photons = s.generate_photons(5)
    assert '`polarization` must be' in str(e.value)

def test_poisson_process():
    '''Do some consistency checks for the Poisson process.

    It turns out that this is hard to test properly, without reimplemention the
    scipy version.
    '''
    p = poisson_process(20.)
    times = p(100., 1. * u.cm**2)
    assert (len(times) > 1500) and (len(times) < 2500)
    assert (times[-1] > 99.) and (times[-1] < 100.)

    times = p(100., 10. * u.cm**2)
    assert (len(times) > 18000) and (len(times) < 22000)
    assert (times[-1] > 99.) and (times[-1] < 100.)

def test_poisson_input():
    '''Input must be a scalar.'''
    with pytest.raises(ValueError) as e:
        p = poisson_process(np.arange(5))
    assert 'must be scalar' in str(e.value)

def test_Aeff():
    '''Check that a higher effective area leads to more counts.'''
    a = RectangleAperture(zoom=2)
    s = Source(flux=50, geomarea=a.area)
    photons = s.generate_photons(5.)
    assert len(photons) == 40

def test_disk_radius():
    pos = SkyCoord(12. * u.degree, -23.*u.degree)
    s = DiskSource(coords=pos, a_inner=50.* u.arcsec,
                   a_outer=10. * u.arcmin)

    photons = s.generate_photons(1e4)
    d = pos.separation(SkyCoord(photons['ra'], photons['dec'], unit='deg'))
    assert np.max(d.arcmin <= 10.)
    assert np.min(d.arcmin >= 0.8)

@pytest.mark.parametrize("diskclass,diskpar",
                         [(DiskSource, {'a_outer': 30. * u.arcmin}),
                          (SphericalDiskSource, {'a_outer': 30. * u.arcmin}),
                          (GaussSource, {'sigma': 1.5 * u.deg})])
def test_disk_distribution(diskclass, diskpar):
    '''This is a separate test from test_disk_radius, because it's a simpler
    to write if we don't have to worry about the inner hole.

    For the test itself: The results should be poisson distributed (or, for large
    numbers this will be almost normal).
    That makes testing it a little awkard in a short run time, thus the limits are
    fairly loose.

    This test is run for several extended sources, incl Gaussian. Stirctly speaking
    it should fail for a Gaussian distribution, but if the sigma is large enough it
    will pass a loose test (and still fail if things to catastrophically wrong,
    e.g. some test circles are outside the source).
    '''

    s = diskclass(coords=SkyCoord(213., -10., unit=u.deg), **diskpar)
    photons = s.generate_photons(1e5)

    n = np.empty(10)
    for i in range(len(n)):
        circ = SkyCoord((213. +  np.random.uniform(-0.1, .1)) * u.degree,
                       (- 10. + np.random.uniform(-0.1, .1))*u.degree)
        d = circ.separation(SkyCoord(photons['ra'], photons['dec'], unit='deg'))
        n[i] = (d < 5. * u.arcmin).sum()
    s, p = normaltest(n)
    # assert a p value here that is soo small that it's never going to be hit
    # by chance.
    assert p > .05
